# SPDX-FileCopyrightText: 2024 Helge
#
# SPDX-License-Identifier: MIT

from unittest.mock import AsyncMock, MagicMock

from . import ActivityRunner


async def test_run_for_modifier():
    app = AsyncMock(actor_uri="https://actor.example/uri", application_name="app")
    sender = AsyncMock(init_create_note=MagicMock())

    app.fetch_activity = AsyncMock(return_value="test")

    runner = ActivityRunner(activity_sender=sender, applications=[app], wait_time=0.1)

    result = await runner.run_for_modifier(lambda x: x)

    app.fetch_activity.assert_awaited_once()

    assert result["app"] == "test"


async def test_run_for_modifier_with_sequence():
    app = AsyncMock(actor_uri="https://actor.example/uri", application_name="app")
    sender = AsyncMock(init_create_note=MagicMock())

    app.fetch_activity = AsyncMock(side_effect=[None, "test"])

    runner = ActivityRunner(activity_sender=sender, applications=[app], wait_time=0.1)

    result = await runner.run_for_modifier(lambda x: x)

    assert result["app"] == "test"



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from unittest.mock import AsyncMock
from contextlib import asynccontextmanager
from fediverse_pasture.types import ApplicationAdapterForActor

from .verify_actor import VerifyActorRunner, format_verify_actor_result


async def test_verify_actor():
    session = AsyncMock()
    runner = VerifyActorRunner(
        application=ApplicationAdapterForActor(
            actor_uri="about:actor", application_name="name"
        ),
        session=session,
    )

    @asynccontextmanager
    async def post_response(*args, **kwargs):
        response = AsyncMock()
        response.json.return_value = {"result": "result", "messages": "messages"}
        yield response

    session.post = post_response

    result = await runner.run()

    assert result == {
        "verify_actor_table": "result",
        "messages": "messages",
        "application_name": "name",
    }


def test_format():
    entry_data = {
        "alice": {"get_actor": False, "post_inbox": False},
        "bob": {"get_actor": True, "post_inbox": True},
        "claire": {"get_actor": True, "post_inbox": True},
        "dean": {"get_actor": False, "post_inbox": False},
        "emily": {"get_actor": True, "post_inbox": True},
        "frank": {"get_actor": True, "post_inbox": True},
    }

    result = [
        {
            "application_name": "test",
            "verify_actor_table": entry_data,
            "messages": "bla",
        }
    ]

    markdown = format_verify_actor_result(result)

    print()
    print()
    print(markdown)



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import logging
import json
import click
import os

from quart import Quart, request, render_template, redirect

from .data_provider import DataProvider
from .types import Message
from .server import BlueprintForActors
from .server.parser import restrictions_for_actors_from_args
from .server.verify_actor import ActorVerifier, remote_urls

from .version import __version__

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


def get_app():
    """Generates the app to be used with an application server

    Usage:

    ```bash
    VERIFY_ACTOR_DOMAIN=my_domain uvicorn --factory fediverse_pasture.verify_actor:get_app
    ```
    """

    return build_app(
        False, os.environ.get("VERIFY_ACTOR_DOMAIN", "localhost:8000"), True
    )


def build_app(
    only_generate_config: bool, domain: str, https: bool, timeout: float
) -> Quart:
    dp = DataProvider.generate_and_load(only_generate_config)
    app = Quart(__name__)

    actor_list = dp.possible_actors + [dp.one_actor]
    app.register_blueprint(BlueprintForActors(actor_list).blueprint)
    actor_names = ["alice", "bob", "claire", "dean", "emily", "frank"]

    async def result_for_actor_uri(actor_uri):
        if not actor_uri:
            return redirect("/")

        message = Message()
        message.add(f"Got Actor Uri {actor_uri}")

        verifier = ActorVerifier(
            actor_list=actor_list,
            remote_uri=actor_uri,
            message=message,
            domain=domain,
            timeout=timeout,
        )

        if https:
            verifier.scheme = "https"

        has_restriction, only = restrictions_for_actors_from_args(
            actor_names, request.args
        )

        result = await verifier.verify(only=only)

        if "json" in request.headers.get("accept"):
            return {
                "result": result,
                "messages": json.dumps(message.response, indent=2),
            }

        warnings = []

        if "alice" in result and result["alice"]["post_inbox"]:
            result["alice"]["warning"] = True
            warnings.append(
                "Alice should not be able to post to the inbox without \
signing her request."
            )

        if "bob" in result and not result["bob"]["get_actor"]:
            result["bob"]["warning"] = True
            warnings.append("Bob should be able to retrieve the actor")
        if "claire" in result and not result["claire"]["get_actor"]:
            result["claire"]["warning"] = True
            warnings.append("Claire should be able to retrieve the actor")

        if "bob" in result and not result["bob"]["post_inbox"]:
            result["bob"]["warning"] = True
            warnings.append("Bob should be able to post to the inbox")
        if "claire" in result and not result["claire"]["post_inbox"]:
            result["claire"]["warning"] = True
            warnings.append(
                """Claire should be able to post to the inbox.
If bob can successfully post to the inbox, this is likely due to using
an unsigned request to retrieve the public key."""
            )

        if not result["webfinger"]:
            warnings.append("""Webfinger result does not match actor URI""")

        return await render_template(
            "verify_actor_result.html.j2",
            messages=json.dumps(message.response, indent=2),
            result={key: value for key, value in result.items() if key != "webfinger"},
            actor_uri=actor_uri,
            warnings=warnings,
            has_warnings=len(warnings) > 0,
            test_actors=actor_names,
            only=only,
            has_restriction=has_restriction,
        )

    @app.get("/static/styles.css")
    async def stylesheet():
        return await render_template("styles.css"), 200, {"content-type": "text/css"}

    @app.get("/static/icon.svg")
    async def icon():
        return (
            await render_template("FediverseLogo.svg"),
            200,
            {"content-type": "image/svg+xml"},
        )

    @app.get("/")
    async def index():
        actor_uri = request.args.get("actor_uri")

        if actor_uri:
            actor_uri = actor_uri.strip()
            return await result_for_actor_uri(actor_uri)

        return await render_template(
            "index.html.j2",
            remote_urls=remote_urls,
            test_actors=actor_names,
        )

    @app.post("/")
    async def verify():
        form_data = await request.form
        actor_uri = form_data.get("actor_uri").strip()

        return await result_for_actor_uri(actor_uri)

    return app


@click.command()
@click.option("--only_generate_config", default=False, is_flag=True)
@click.option("--reload", default=False, is_flag=True)
@click.option(
    "--https",
    default=False,
    is_flag=True,
    help="Flag indicates that the server runs behind https",
)
@click.option("--port", default=2909, help="port to run on")
@click.option(
    "--domain", default="localhost:2909", help="domain the service is running on"
)
@click.option(
    "--timeout",
    default=20,
    help="Timeout for requests when verifying the actor in seconds",
)
def verify_actor(only_generate_config, reload, https, port, domain, timeout):
    print(f"fediverse pasture one_actor {__version__}")

    app = build_app(only_generate_config, domain, https, timeout)
    app.run(port=port, host="0.0.0.0", use_reloader=reload)


if __name__ == "__main__":
    verify_actor()



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT


from quart import Blueprint
from PIL import Image
from random import randint
from io import BytesIO

image_provider_blueprint = Blueprint("image_provider", __name__)
"""Provides the capability to generate randomly colored 40 by 40 pixel images in various formats"""


def determine_format(filename):
    if filename.endswith("jpg"):
        return "JPEG", "image/jpeg"
    if filename.endswith("eps"):
        return "EPS", "image/eps"
    if filename.endswith("gif"):
        return "GIF", "image/gif"
    if filename.endswith("tiff"):
        return "TIFF", "image/tiff"
    if filename.endswith("webp"):
        return "WEBP", "image/webp"
    return "PNG", "image/png"


@image_provider_blueprint.get("/<filename>")
async def get_filename(filename):
    """Returns a random image. If filename ends with jpg, eps, gif, tiff, or webp an image of that type is returned otherwise png"""
    image_format, content_type = determine_format(filename)

    b = BytesIO()
    image = Image.new(
        "RGB", (40, 40), (randint(0, 255), randint(0, 255), randint(0, 255))
    )
    image.save(b, format=image_format)

    return b.getvalue(), 200, {"content-type": content_type}



# SPDX-FileCopyrightText: 2024 Helge
#
# SPDX-License-Identifier: MIT

import asyncio
import click

from .helper import activity_runner


@click.group
def group(): ...


async def send_message(application, verify=False):
    async with activity_runner(application) as runner:
        runner.skip_fetch = not verify

        return await runner.run_for_modifier(lambda x: {**x, "content": "Hello World"})


@group.command("hello", help="Sends hello world to the application")
@click.argument("application")
def message(application):
    asyncio.run(send_message(application))


@group.command("verify", help="Verifies the application works correctly")
@click.argument("application")
def verify(application):
    result = asyncio.run(send_message(application, verify=True))

    if not result:
        click.echo("Failed to retrieve message")
        exit(1)


if __name__ == "__main__":
    group()



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from tortoise.models import Model
from tortoise import fields


class TestRecord(Model):
    id = fields.IntField(primary_key=True)
    test_name = fields.CharField(max_length=255)
    application_name = fields.CharField(max_length=255)

    data = fields.JSONField()

    class Meta:
        unique_together = (("test_name", "application_name"),)



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import base64
import aiohttp
import json
from typing import Tuple

from bovine.jsonld import with_external_context
from bovine.utils import parse_gmt, check_max_offset_now
from bovine.crypto.signature import Signature
from bovine.crypto.helper import sign_message
from bovine.crypto.http_signature import build_signature

from fediverse_pasture.types import Message


def find_with_item(dict_list, key_id):
    """Given a list of dictionaries, finds the dictionary with
    id = key_id"""
    for key in dict_list:
        if key.get("id") == key_id:
            return key
    return None


def public_key_owner_from_dict(
    actor: dict, key_id: str
) -> Tuple[str | None, str | None]:
    """Given an actor and key_id returns the public_key and the owner. This method directly checks the key `publicKey`"""

    public_key_data = actor.get("publicKey", {})

    if isinstance(public_key_data, list):
        if len(public_key_data) == 1:
            public_key_data = public_key_data[0]
        else:
            public_key_data = find_with_item(public_key_data, key_id)

    if not public_key_data:
        return None, None

    public_key = public_key_data.get("publicKeyPem")
    owner = public_key_data.get("owner")

    return public_key, owner


def actor_object_to_public_key(
    actor: dict, key_id: str
) -> Tuple[str | None, str | None]:
    """As [public_key_owner_from_dict][fediverse_pasture.server.utils.public_key_owner_from_dict], but applies context normalization first, then checks without out."""
    public_key, owner = public_key_owner_from_dict(with_external_context(actor), key_id)

    if public_key and owner:
        return public_key, owner

    return public_key_owner_from_dict(actor, key_id)


def build_message_signature(request, method):
    message = Message()
    message.add(f"Got {method} request")
    message.add("With headers: " + str(request.headers))

    signature_header = request.headers.get("signature")
    if not signature_header:
        return None, None, message.error("Signature header is missing")

    message.add(f"Signature header '{signature_header}'")

    try:
        signature = Signature.from_signature_header(signature_header)
    except Exception as e:
        return (
            None,
            None,
            message.error(
                [
                    "Failed to parse signature",
                    repr(e),
                ]
            ),
        )

    message.add(f"""Got fields {", ".join(signature.fields)}""")

    return message, signature, None


def validate_basic_signature_fields(message, request, signature, digest=None):
    if "(request-target)" not in signature.fields:
        return message.error("(request-target) must be a signature field")

    if "host" not in signature.fields:
        return message.error("host must be a signature field")

    if "date" not in signature.fields:
        return message.error("date must be a signature field")

    if digest:
        if "digest" not in signature.fields:
            return message.error("digest must be a signature field")

        if request.headers.get("digest")[5:] != digest[5:]:
            return message.error("Digests do not match")

    date_header = request.headers.get("date")
    message.add(f"Got date header {date_header}")

    try:
        date_parsed = parse_gmt(date_header)
    except Exception as e:
        return message.error(["Failed to parse date", repr(e)])

    if not check_max_offset_now(date_parsed):
        return message.error(["Date not within the last 5 minutes"])

    return None


async def retrieve_public_key(message, key_id, request):
    if key_id == "about:inline":
        try:
            public_key = base64.standard_b64decode(
                request.headers.get("X-Public-Key")
            ).decode("utf-8")
        except Exception:
            public_key = None
        if not public_key:
            return None, message.error(
                "Please set the header 'X-Public-Key' to provide a public key"
            )
    else:
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(key_id) as response:
                    if "json" in response.headers.get("content-type"):
                        data = await response.json()
                        message.add("Got json response " + json.dumps(data, indent=2))
                        public_key, _ = actor_object_to_public_key(data, key_id)
                    else:
                        public_key = await response.text()
        except Exception as e:
            return (
                None,
                message.error(
                    [
                        "Failed to fetch public key",
                        "Use about:inline to include a public key in a header",
                        "    about:inline is expected to be base64 encoded",
                        "or make sure keyId can be resolved to a public key",
                        repr(e),
                    ]
                ),
            )
    if not public_key:
        return None, message.error("Something went wrong when retrieving public key")
    message.add(f"""Got public key "{public_key}" """)

    return public_key, None


def build_http_signature(message, signature, request, method):
    header_fields = [x for x in signature.fields if x[0] != "("]

    http_signature = build_signature(request.headers.get("host"), method, request.path)

    for x in header_fields:
        if x != "host":
            http_signature.with_field(x, request.headers.get(x))

    message.add(f"""Message to sign "{http_signature.build_message()}" """)
    message.add(f"Got key id {signature.key_id}")

    return http_signature


def validate_signature(message, request, public_key, signature, http_signature):
    try:
        result = http_signature.verify(
            public_key=public_key, signature=signature.signature
        )
    except Exception as e:
        return (
            message.error(
                [
                    "Something went wrong when verifying signature",
                    repr(e),
                ]
            ),
            401,
        )

    if not result:
        private_key = request.headers.get("X-Private-Key")

        if private_key is None:
            message.add(
                "Set X-Private-Key to your base64 encoded private key\
                        for expected signature",
            )
        else:
            message.add("Expected signature:")
            message.add(
                sign_message(
                    private_key=base64.standard_b64decode(private_key).decode("utf-8"),
                    message=http_signature.build_message(),
                )
            )

        return message.error("Invalid signature")
    else:
        message.add("SUCCESS!!!")



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import json
import textwrap
from dataclasses import dataclass, field
from typing import Dict, Set, List, Callable
import logging

logger = logging.getLogger(__name__)


@dataclass
class Entry:
    """Represents an entry to generate support tables from."""

    entry: Dict[str, Dict] = field(
        metadata={
            "description": "Dictionary indexed by application with values the result"
        }
    )

    @property
    def applications(self) -> Set[str]:
        """Returns the set of application names"""
        return set(self.entry.keys())

    @property
    def activity(self) -> Dict:
        result = self.entry.get("activity")
        if result is None:
            raise ValueError("No activity data found in entry")
        return result

    @property
    def object(self):
        activity = self.activity
        obj = activity.get("object", {})
        return {"@context": activity["@context"], **obj}

    def present_for(self, application):
        return application in self.entry

    def as_tabs(self, apps: List[str]) -> List[str]:
        """Renders the data for each application as tabbed markdown.

        If no data is present "no result" is shown.

        :param apps: List of applications to display"""
        lines = []
        for application_name in apps:
            lines.append(f"""=== "{application_name}"\n""")

            if application_name in self.entry:
                text = f"""```json title="{application_name}"\n"""
                text += json.dumps(self.entry[application_name], indent=2)
                text += "\n```\n"
                lines += textwrap.indent(text, " " * 4).split("\n")
            else:
                lines.append("    no result")

            lines += [""]

        return [x + "\n" for x in lines]

    def as_grid(self, apps: List[str]) -> List[str]:
        """Renders the data for each application as a grid.

        If no data is present "no result" is shown.

        :param apps: List of applications to display"""
        lines = ["""<div class="grid" markdown>\n"""]
        for application_name in apps:
            if application_name in self.entry:
                text = f"""```json title="{application_name}"\n"""
                text += json.dumps(self.entry[application_name], indent=2)
                text += "\n```\n"
                lines += textwrap.indent(text, " " * 0).split("\n")
            else:
                lines.append("no result")

            lines += [""]

        lines.append("</div>")

        return [x + "\n" for x in lines]

    def apply_to(
        self, application: str, function: Callable[[Dict], List[str]]
    ) -> List[str]:
        """Applies the function to the entry for the given application.

        :param application: The application name
        :param function: extractor for the desired data"""
        try:
            data = self.entry.get(application)
            if data is None:
                logger.warning(f"No data found for application {application}")
                return ["-"]

            return function(data)
        except Exception as e:
            logger.exception(e)
            return ["-"]

    @staticmethod
    def from_result_list(result):
        val = {x["application_name"]: x for x in result}
        for x in val:
            del val[x]["application_name"]
        return Entry(val)



# SPDX-FileCopyrightText: 2023-2024 Helge
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
import logging
import json
import aiohttp

from bovine.clients.bearer import BearerAuthClient

from fediverse_pasture.types import ApplicationAdapterForLastActivity

logger = logging.getLogger(__name__)


@dataclass
class MastodonApplication:
    domain: str
    access_token: str
    username: str
    client: BearerAuthClient | None = None
    actor_uri: str | None = None

    def determine_actor_uri(self):
        if self.actor_uri:
            return self.actor_uri
        return f"http://{self.domain}/users/{self.username}"

    async def public_items(self):
        response = await self.client.get(
            f"http://{self.domain}/api/v1/timelines/public"
        )
        public_timeline = await response.json()
        logger.debug(json.dumps(public_timeline, indent=2))

        return public_timeline

    async def top_public_with_object_id(self, object_id: str) -> dict | None:
        for item in await self.public_items():
            if object_id in json.dumps(item):
                return item

        return None

    def last_activity(
        self, session: aiohttp.ClientSession, application_name: str = "mastodon"
    ) -> ApplicationAdapterForLastActivity:
        self.client = BearerAuthClient(session, self.access_token)

        return ApplicationAdapterForLastActivity(
            actor_uri=self.determine_actor_uri(),
            fetch_activity=self.top_public_with_object_id,
            application_name=application_name,
        )



from dataclasses import dataclass

from fediverse_pasture.types import MessageModifier


def get_inputs_value(input_name: str, input_number: int) -> dict:
    try:
        import fediverse_pasture_inputs
    except ImportError:
        raise ImportError(
            "The 'fediverse_pasture_inputs' package is required to use input modifiers."
        )
    inputs = fediverse_pasture_inputs.available[input_name]

    return inputs.examples[input_number]


@dataclass
class ModifierBuilder:
    text: str | None = None
    input_name: str | None = None
    input_number: int | None = None

    def __post_init__(self):
        if (self.input_name and self.input_number is None) or (
            not self.input_name and self.input_number
        ):
            raise ValueError(
                "Both input_name and input_number must be specified together."
            )

        if self.text and self.input_name and self.input_number:
            raise ValueError(
                "Cannot specify text, input_name, and input_number at the same time."
            )

    def build(self) -> MessageModifier:
        if self.text:
            return lambda x: {**x, "content": self.text}

        if self.input_name and self.input_number is not None:
            info = get_inputs_value(self.input_name, self.input_number)

            return lambda x: {**x, "content": "text", **info}

        return lambda x: {
            **x,
            "content": "text",
        }



from .id_maker import create_make_id_for_actor_id


def test_create_make_id_for_actor_id():
    actor_id = "https://actor.test/actor_id"
    make_id = create_make_id_for_actor_id(actor_id)

    assert make_id().startswith("https://actor.test/object/")



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import aiohttp
from typing import Callable, Awaitable
from bovine.clients import lookup_uri_with_webfinger

from fediverse_pasture.types import (
    ApplicationAdapterForActor,
    ApplicationAdapterForLastActivity,
)


from .mastodon import MastodonApplication
from .misskey import MisskeyApplication
from .firefish import FirefishApplication
from .bovine import BovineApplication


async def actor_for_application(
    account_uri: str, application_name: str, session: aiohttp.ClientSession
) -> ApplicationAdapterForActor:
    """Creates a ApplicationAdapterForActor

    :param account_uri: The acct uri, e.g. `acct:user@domain`
    :param application_name: The name of the application
    :param session: the aiohttp ClientSession
    """

    domain = account_uri.split("@")[1]

    actor_uri, _ = await lookup_uri_with_webfinger(
        session, account_uri, f"http://{domain}"
    )

    if not actor_uri:
        raise ValueError(f"Actor not found with URI {account_uri}")

    return ApplicationAdapterForActor(
        actor_uri=actor_uri, application_name=application_name
    )


async def activity_for_mastodon(
    domain: str,
    username: str,
    access_token: str,
    session: aiohttp.ClientSession,
    application_name: str = "mastodon",
    determine_actor_uri: bool = False,
) -> ApplicationAdapterForLastActivity:
    """Creates a ApplicationAdapterForLastActivity object for connecting to
    mastodon. Example usage:

    ```python
    mastodon = await activity_for_mastodon("mastodon_web", "bob", "xxx", session)
    ```
    """

    actor_uri = None

    if determine_actor_uri:
        actor_uri, _ = await lookup_uri_with_webfinger(
            session, f"acct:{username}@{domain}", f"http://{domain}"
        )

    mastodon = MastodonApplication(
        domain=domain, access_token=access_token, username=username, actor_uri=actor_uri
    )

    return mastodon.last_activity(session, application_name=application_name)


async def activity_for_firefish(
    domain: str, username: str, session: aiohttp.ClientSession
) -> ApplicationAdapterForLastActivity:
    """Creates a ApplicationAdapterForLastActivity object for connecting to
    firefish. Example usage:

    ```python
    firefish = await activity_for_firefish("firefish_web", "admin", session)
    ```
    """
    firefish = FirefishApplication(domain=domain, username=username)

    return await firefish.last_activity(session)


def activity_for_mastodon_provider(
    domain: str,
    username: str,
    access_token: str,
    application_name: str = "mastodon",
    determine_actor_uri: bool = False,
) -> Callable[[aiohttp.ClientSession], Awaitable[ApplicationAdapterForLastActivity]]:
    def func(session):
        return activity_for_mastodon(
            domain,
            username,
            access_token,
            session,
            application_name=application_name,
            determine_actor_uri=determine_actor_uri,
        )

    return func


def activity_for_misskey_provider(
    domain: str,
    username: str,
) -> Callable[[aiohttp.ClientSession], Awaitable[ApplicationAdapterForLastActivity]]:
    def func(session: aiohttp.ClientSession):
        async def misskey():
            app = MisskeyApplication(domain, username, session=session)

            return await app.last_activity()

        return misskey()

    return func


def activity_for_bovine_provider(domain, username, secret):
    app = BovineApplication(domain=domain, username=username, secret=secret)

    def func(session):
        return app.last_activity(session)

    return func


async def firefish_app(session):
    return await activity_for_firefish("firefish_web", "admin", session)


async def mitra_app(session):
    async with session.post(
        "http://mitra/oauth/token",
        data={"grant_type": "password", "username": "admin", "password": "password"},
    ) as response:
        data = await response.json()

    return await activity_for_mastodon(
        "mitra", "admin", data["access_token"], session, application_name="mitra"
    )


async def hubzilla_app(session):
    app = await actor_for_application("admin@hubzilla", "hubzilla", session)
    app.application_name = "hubzilla"
    return app


async def custom_app(session):
    from custom import app_from_session

    return await app_from_session(session)


mastodon_likes = {
    application_name: activity_for_mastodon_provider(
        application_name, username, "token", application_name=application_name
    )
    for application_name, username in [
        ("mastodon", "hippo"),
        ("gotosocial", "cookie"),
        ("akkoma", "witch"),
        ("pleroma", "full"),
    ]
}


app_name_to_coroutine = {
    **mastodon_likes,
    "bovine": activity_for_bovine_provider(
        "bovine", "vanilla", "z3u2Yxcowsarethebestcowsarethebestcowsarethebest"
    ),
    "misskey": activity_for_misskey_provider("misskey", "kitty"),
    "mitra": mitra_app,
    "firefish": firefish_app,
    "hubzilla": hubzilla_app,
    "sharkey": activity_for_mastodon_provider(
        "sharkey",
        "willy",
        "token",
        determine_actor_uri=True,
        application_name="sharkey",
    ),
    "custom": custom_app,
}
"""Used to lookup the coroutine to get the application object"""



import pytest

from .modifier import ModifierBuilder


@pytest.mark.parametrize(
    ["text", "input_name", "input_number"],
    [("text", "name", 1), (None, None, 1), (None, "name", None)],
)
def test_invalid_arguments(text, input_name, input_number):
    with pytest.raises(ValueError):
        ModifierBuilder(text=text, input_name=input_name, input_number=input_number)


def test_build_with_text():
    builder = ModifierBuilder(text="Hello, World!")
    modifier = builder.build()

    assert modifier({}) == {"content": "Hello, World!"}


def test_with_input():
    builder = ModifierBuilder(input_name="hashtags", input_number=1)
    modifier = builder.build()

    assert modifier({}) == {
        "content": "text",
        "tag": {"name": "nohash", "type": "Hashtag"},
    }


def test_with_input_zero_value():
    builder = ModifierBuilder(input_name="hashtags", input_number=0)
    modifier = builder.build()

    assert modifier({}) == {
        "content": "text",
        "tag": {"name": "#test", "type": "Hashtag"},
    }



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from .result_store import with_store
from .entry import Entry


async def test_with_store():
    async with with_store(db_url="sqlite://:memory:") as store:
        await store.add_result("test", "app1", {})
        await store.add_result("test", "app1", {"a": "b"})
        await store.add_result("test", "app2", {})
        await store.add_result("other", "app1", {})

        result = await store.results_for_test("test")

        assert len(result) == 2


async def test_delete():
    async with with_store(db_url="sqlite://:memory:") as store:
        await store.add_result("test", "app1", {})
        await store.delete_record("test", "app1")

        result = await store.results_for_test("test")

        assert len(result) == 0


async def test_with_store_to_entry():
    async with with_store(db_url="sqlite://:memory:") as store:
        await store.add_result("test", "app1", {})
        await store.add_result("test", "app1", {"a": "b"})
        await store.add_result("test", "app2", {})

        result = await store.entry_for_test("test")

        assert isinstance(result, Entry)

        print(result.entry)


async def test_with_store_as_list():
    async with with_store(db_url="sqlite://:memory:") as store:
        await store.add_result("test1", "app1", {"a": "b"})
        await store.add_result("test1", "app2", {"a": 1})

        await store.add_result("test2", "app1", {"a": "b"})
        await store.add_result("test2", "app2", {"a": 2})

        result = await store.as_list()

        assert isinstance(result, list)

        assert len(result) == 4



# SPDX-FileCopyrightText: 2024 Helge
#
# SPDX-License-Identifier: MIT

import aiohttp

import pytest
from unittest.mock import MagicMock, AsyncMock
from .verify_actor import ActorVerifier, can_be_resolved, remote_urls


@pytest.mark.parametrize(
    ["uri", "resolvable"],
    [
        ("http://localhost/actor", True),
        ("https://remote.example", True),
        ("@remote@something.example", False),
    ],
)
def test_can_be_resolved(uri, resolvable):
    assert can_be_resolved(uri) == resolvable


@pytest.mark.parametrize(
    "remote_uri", ["https://remote.example/actor", "http://localhost/actor"]
)
async def test_determine_remote_actor_uri(remote_uri):
    verifier = ActorVerifier([], remote_uri, "http://localhost", MagicMock())

    await verifier.determine_remote_actor_uri(AsyncMock())

    assert verifier.remote_actor_uri == remote_uri


async def test_determine_remote_remote_actor_uri_invalid():
    verifier = ActorVerifier([], "invalid", "http://localhost", MagicMock())

    await verifier.determine_remote_actor_uri(AsyncMock())


@pytest.fixture
async def client_session():
    async with aiohttp.ClientSession() as session:
        yield session


@pytest.mark.parametrize(["key", "url"], remote_urls.items())
async def test_remote_urls(client_session, key, url):
    async with client_session.get(url) as response:
        assert response.status == 200



# SPDX-FileCopyrightText: 2023, 2024 Helge
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from datetime import datetime
import logging
import aiohttp

from bovine import BovineClient

from fediverse_pasture.types import ApplicationAdapterForLastActivity

logger = logging.getLogger(__name__)


@dataclass
class BovineApplication:
    domain: str
    username: str
    secret: str
    actor_uri: str | None = None
    client: BovineClient | None = None

    async def fetch_activity(self, published):
        try:
            inbox = self.client.inbox()

            result = await inbox.__anext__()
            parsed_published = datetime.fromisoformat(
                result["object"]["published"].removesuffix("Z")
            )
            if parsed_published == published:
                return result.get("object")
        except Exception as e:
            print("Something went wrong with", repr(e))
        return None

    async def last_activity(
        self, session: aiohttp.ClientSession
    ) -> ApplicationAdapterForLastActivity:
        self.client = BovineClient(
            domain=f"http://{self.domain}",
            secret=self.secret,
        )
        await self.client.init(session=session)

        return ApplicationAdapterForLastActivity(
            actor_uri=self.client.actor_id,
            fetch_activity=self.fetch_activity,
            application_name=self.domain,
        )



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
import tomli_w
import tomllib
from typing import Tuple, List

from bovine import BovineActor
from bovine.activitystreams import Actor

from .generate import generate_one_actor, generate_possible_actors
from .models import ActorData


@dataclass
class DataProvider:
    """Provides the data about actors to various applications"""

    one_actor: ActorData = field(
        metadata={"description": "User in one actor and as an application actor"}
    )
    possible_actors: List[ActorData] = field(
        metadata={"description": "The list of actors used for verify_actor"}
    )

    def save(self):
        data = {
            "one_actor": self.one_actor.model_dump(exclude_none=True),
            "possible_actors": [
                x.model_dump(exclude_none=True) for x in self.possible_actors
            ],
        }

        with open("data.toml", "wb") as fp:
            tomli_w.dump(data, fp)

    @staticmethod
    def load():
        with open("data.toml", "rb") as fp:
            data = tomllib.load(fp)

        one_actor = ActorData(**(data["one_actor"]))
        possible_actors = [ActorData(**x) for x in data["possible_actors"]]

        return DataProvider(one_actor=one_actor, possible_actors=possible_actors)

    @staticmethod
    def generate(with_possible_actors=True):
        one_actor = generate_one_actor()
        if with_possible_actors:
            possible_actors = generate_possible_actors()
        else:
            possible_actors = []
        return DataProvider(one_actor=one_actor, possible_actors=possible_actors)

    @staticmethod
    def generate_and_load(only_generate_config):
        if only_generate_config:
            dp = DataProvider.generate()
            dp.save()
            exit(0)

        try:
            dp = DataProvider.load()
        except Exception:
            dp = DataProvider.generate()
            dp.save()

        return dp


def bovine_actor_for_actor_data(
    actor_id: str, data: ActorData
) -> Tuple[BovineActor, Actor]:
    """Builds the corresponding bovine actor and actor data

    :param actor_id: The actor id to be used, e.g. `https://pasture_one_actor/actor`
    :param data: The actor data
    :return: A tuple, where the first object is the Actor performing actions, and the second one the data object.
    """
    key_pair = data.key_pairs[0]
    bovine_actor = BovineActor(
        actor_id=actor_id,
        public_key_url=f"{actor_id}#{key_pair.name}",
        secret=key_pair.private,
    )

    actor_object = Actor(
        id=actor_id,
        preferred_username=data.user_part,
        name="Test Actor",
        inbox=f"{actor_id}/inbox",
        outbox=f"{actor_id}/outbox",
        public_key=key_pair.public,
        public_key_name=key_pair.name,
        summary=data.summary,
    )

    return bovine_actor, actor_object



# SPDX-FileCopyrightText: 2024 Helge
#
# SPDX-License-Identifier: MIT

import pytest

from quart import Quart

from . import assets_blueprint_for_directory


@pytest.fixture
def test_client():
    app = Quart(__name__)
    app.register_blueprint(assets_blueprint_for_directory("assets"))

    yield app.test_client()


async def test_not_found(test_client):
    response = await test_client.get("/assets/unknown")
    assert response.status_code == 404


async def test_json(test_client):
    response = await test_client.get("/assets/test.json")

    assert response.status_code == 200
    assert response.headers.get("content-type") == "application/json"


async def test_activitypub(test_client):
    response = await test_client.get("/assets/test.jsonap")

    assert response.status_code == 200
    assert response.headers.get("content-type") == "application/activity+json"



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from .utils import actor_object_to_public_key

public_key_string = """-----BEGIN PUBLIC KEY-----
CONTENT
-----END PUBLIC KEY-----
"""

base_case_actor = {
    "@context": [
        "https://www.w3.org/ns/activitystreams",
        "https://w3id.org/security/v1",
    ],
    "id": "https://domain.example/actor",
    "type": "Person",
    "preferredUsername": "alice",
    "inbox": "https://domain.example/inbox",
}

key_id = "https://domain.example/actor#main-key"


def test_base_case():
    data = {
        **base_case_actor,
        "publicKey": {
            "id": key_id,
            "owner": "https://domain.example/actor",
            "publicKeyPem": public_key_string,
        },
    }
    public_key, _ = actor_object_to_public_key(data, key_id)

    assert public_key == public_key_string


def test_base_case_wrong_context():
    data = {
        **base_case_actor,
        "@context": "https://www.w3.org/ns/activitystreams",
        "publicKey": {
            "id": key_id,
            "owner": "https://domain.example/actor",
            "publicKeyPem": public_key_string,
        },
    }
    public_key, _ = actor_object_to_public_key(data, key_id)

    assert public_key == public_key_string


def test_base_case_expanded():
    data = {
        **base_case_actor,
        "@context": "https://www.w3.org/ns/activitystreams",
        "https://w3id.org/security#publicKey": {
            "id": key_id,
            "https://w3id.org/security#owner": {"id": "https://domain.example/actor"},
            "https://w3id.org/security#publicKeyPem": public_key_string,
        },
    }
    public_key, _ = actor_object_to_public_key(data, key_id)

    assert public_key == public_key_string


def test_single_item_list():
    data = {
        **base_case_actor,
        "publicKey": [
            {
                "id": key_id,
                "owner": "https://domain.example/actor",
                "publicKeyPem": public_key_string,
            }
        ],
    }
    public_key, _ = actor_object_to_public_key(data, key_id)

    assert public_key == public_key_string


def test_multi_item_list():
    data = {
        **base_case_actor,
        "publicKey": [
            {
                "id": key_id,
                "owner": "https://domain.example/actor",
                "publicKeyPem": public_key_string,
            },
            {
                "id": "https://domain.example/actor#other-key",
                "owner": "https://domain.example/actor",
                "publicKeyPem": "xxxx",
            },
        ],
    }
    public_key, _ = actor_object_to_public_key(data, key_id)

    assert public_key == public_key_string


def test_multi_item_list_no_result():
    data = {
        **base_case_actor,
        "publicKey": [
            {
                "id": "https://domain.example/actor#wrong-key",
                "owner": "https://domain.example/actor",
                "publicKeyPem": public_key_string,
            },
            {
                "id": "https://domain.example/actor#other-key",
                "owner": "https://domain.example/actor",
                "publicKeyPem": "xxxx",
            },
        ],
    }
    public_key, _ = actor_object_to_public_key(data, key_id)

    assert public_key is None
